@@ -29,10 +29,21 @@
   "The EasyPG Assistant hooks for transparent file encryption"
   :group 'epa)
 
-(defcustom epa-file-name-regexp "\\.gpg\\'"
-  "Regexp which matches filenames to be encrypted with GnuPG."
+(defun epa-file--file-name-regexp-set (variable value)
+  (set-default variable value)
+  (if (fboundp 'epa-file-name-regexp-update)
+      (epa-file-name-regexp-update)))
+
+(defcustom epa-file-name-regexp "\\.gpg\\(~\\|\\.~[0-9]+~\\)?\\'"
+  "Regexp which matches filenames to be encrypted with GnuPG.
+
+If you set this outside Custom while epa-file is already enabled, you
+have to call `epa-file-name-regexp-update' after setting it to
+properly update file-name-handler-alist.  Setting this through Custom
+does that automatically."
   :type 'regexp
-  :group 'epa-file)
+  :group 'epa-file
+  :set 'epa-file--file-name-regexp-set)
 
 (defcustom epa-file-cache-passphrase-for-symmetric-encryption nil
   "If non-nil, cache passphrase for symmetric encryption."
@@ -44,6 +55,11 @@
   :type 'boolean
   :group 'epa-file)
 
+(defcustom epa-file-select-keys nil
+  "If non-nil, always asks user to select recipients."
+  :type 'boolean
+  :group 'epa-file)
+
 (defvar epa-file-encrypt-to nil
   "*Recipient(s) used for encrypting files.
 May either be a string or a list of strings.")
@@ -60,18 +76,28 @@ May either be a string or a list of strings.")
 			val)
 		  t)))))
 
+;;;###autoload
+(put 'epa-file-encrypt-to 'permanent-local t)
+
 (defvar epa-file-handler
   (cons epa-file-name-regexp 'epa-file-handler))
 
 (defvar epa-file-passphrase-alist nil)
 
-(if (fboundp 'encode-coding-string)
-    (defalias 'epa-file--encode-coding-string 'encode-coding-string)
-  (defalias 'epa-file--encode-coding-string 'identity))
+(eval-and-compile
+  (if (fboundp 'encode-coding-string)
+      (defalias 'epa-file--encode-coding-string 'encode-coding-string)
+    (defalias 'epa-file--encode-coding-string 'identity)))
 
-(if (fboundp 'decode-coding-string)
-    (defalias 'epa-file--decode-coding-string 'decode-coding-string)
-  (defalias 'epa-file--decode-coding-string 'identity))
+(eval-and-compile
+  (if (fboundp 'decode-coding-string)
+      (defalias 'epa-file--decode-coding-string 'decode-coding-string)
+    (defalias 'epa-file--decode-coding-string 'identity)))
+
+(defun epa-file-name-regexp-update ()
+  (interactive)
+  (unless (equal (car epa-file-handler) epa-file-name-regexp)
+    (setcar epa-file-handler epa-file-name-regexp)))
 
 (defun epa-file-passphrase-callback-function (context key-id file)
   (if (and epa-file-cache-passphrase-for-symmetric-encryption
@@ -125,8 +151,6 @@ May either be a string or a list of strings.")
   (barf-if-buffer-read-only)
   (if (and visit (or beg end))
       (error "Attempt to visit less than an entire file"))
-  (if epa-file-inhibit-auto-save
-      (auto-save-mode 0))
   (setq file (expand-file-name file))
   (let ((local-copy (epa-file-run-real-handler #'file-local-copy (list file)))
 	(context (epg-make-context))
@@ -150,6 +174,9 @@ May either be a string or a list of strings.")
 		 (setcdr entry nil))
 	     (signal 'file-error
 		     (cons "Opening input file" (cdr error)))))
+	  (make-local-variable 'epa-file-encrypt-to)
+	  (setq epa-file-encrypt-to
+		(mapcar #'car (epg-context-result-for context 'encrypted-to)))
 	  (if (or beg end)
 	      (setq string (substring string (or beg 0) end)))
 	  (save-excursion
@@ -182,13 +209,19 @@ May either be a string or a list of strings.")
 			      buffer-file-coding-system)))
 	 (context (epg-make-context))
 	 (coding-system-for-write 'binary)
-	 string entry)
+	 string entry
+	 (recipients
+	  (cond
+	   ((listp epa-file-encrypt-to) epa-file-encrypt-to)
+	   ((stringp epa-file-encrypt-to) (list epa-file-encrypt-to)))))
     (epg-context-set-passphrase-callback
      context
      (cons #'epa-file-passphrase-callback-function
 	   file))
     (epg-context-set-progress-callback context
 				       #'epa-progress-callback-function)
+    (epg-context-set-armor context epa-armor)
+    (epg-context-set-textmode context epa-textmode)
     (condition-case error
 	(setq string
 	      (epg-encrypt-string
@@ -197,14 +230,15 @@ May either be a string or a list of strings.")
 		   (epa-file--encode-coding-string start coding-system)
 		 (epa-file--encode-coding-string (buffer-substring start end)
 						 coding-system))
-	       (unless (assoc file epa-file-passphrase-alist)
-		 (epa-select-keys
-		  context
-		  "Select recipents for encryption.
+	       (if (or epa-file-select-keys
+		       (not (local-variable-p 'epa-file-encrypt-to)))
+		   (epa-select-keys
+		    context
+		    "Select recipents for encryption.
 If no one is selected, symmetric encryption will be performed.  "
-		  (cond
-		   ((listp epa-file-encrypt-to) epa-file-encrypt-to)
-		   ((stringp epa-file-encrypt-to) (list epa-file-encrypt-to)))))))
+		    recipients)
+		 (if epa-file-encrypt-to
+		     (epg-list-keys context recipients)))))
       (error
        (if (setq entry (assoc file epa-file-passphrase-alist))
 	   (setcdr entry nil))
@@ -228,6 +262,22 @@ If no one is selected, symmetric encryption will be performed.  "
 	(message "Wrote %s" buffer-file-name))))
 (put 'write-region 'epa-file 'epa-file-write-region)
 
+(defun epa-file-find-file-hook ()
+  (if (and buffer-file-name
+	   (string-match epa-file-name-regexp buffer-file-name)
+	   epa-file-inhibit-auto-save)
+      (auto-save-mode 0)))
+
+(defun epa-file-select-keys ()
+  "Select recipients for encryption."
+  (interactive)
+  (make-local-variable 'epa-file-encrypt-to)
+  (setq epa-file-encrypt-to
+	(epa-select-keys
+	 (epg-make-context)
+	 "Select recipents for encryption.
+If no one is selected, symmetric encryption will be performed.  ")))
+
 ;;;###autoload
 (defun epa-file-enable ()
   (interactive)
@@ -235,6 +285,7 @@ If no one is selected, symmetric encryption will be performed.  "
       (message "`epa-file' already enabled")
     (setq file-name-handler-alist
 	  (cons epa-file-handler file-name-handler-alist))
+    (add-hook 'find-file-hooks 'epa-file-find-file-hook)
     (message "`epa-file' enabled")))
 
 ;;;###autoload
@@ -244,6 +295,7 @@ If no one is selected, symmetric encryption will be performed.  "
       (progn
 	(setq file-name-handler-alist
 	      (delq epa-file-handler file-name-handler-alist))
+	(remove-hook 'find-file-hooks 'epa-file-find-file-hook)
 	(message "`epa-file' disabled"))
     (message "`epa-file' already disabled")))
 
